/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <gnome.h>

#include "synctmr.h"

struct stSyncTmrFunc {
	GtkFunction function;
	gpointer data;
};

struct stSyncTmr {
	GList *events;
	GHashTable *timeouts;
};


struct stSyncTmrEvent {
	guint timeout_handler_id;
	struct stSyncTmr *synctmr;
};

struct stSyncTmr *synctmr_create(void)
{
	struct stSyncTmr *synctmr;
	synctmr = g_new0(struct stSyncTmr, 1);
	synctmr->events = NULL;
	synctmr->timeouts =
	    g_hash_table_new(g_direct_hash, g_direct_equal);
	return synctmr;
}

static void antecipate_timeouts(gpointer key, gpointer value,
				gpointer user_data)
{
	struct stSyncTmrEvent *synctmrev = value;
	struct stSyncTmr *synctmr = synctmrev->synctmr;
	struct stSyncTmrFunc *synctmrfunc;
	GList *firstelem;

	gtk_timeout_remove(synctmrev->timeout_handler_id);

	firstelem = g_list_first(synctmr->events);
	synctmrfunc = (struct stSyncTmrFunc *) firstelem->data;
	synctmr->events = g_list_remove_link(synctmr->events, firstelem);
	g_list_free_1(firstelem);
	g_free(synctmrev);

	/* run timeout, ignore return value (we don't repeat timeouts) */
	(*synctmrfunc->function) (synctmrfunc->data);
	g_free(synctmrfunc);
}

void synctmr_destroy(struct stSyncTmr *synctmr)
{
	g_hash_table_foreach(synctmr->timeouts, antecipate_timeouts, NULL);
	g_hash_table_destroy(synctmr->timeouts);
	g_free(synctmr);
}

static gint timeout_happened(gpointer data)
{
	struct stSyncTmrEvent *synctmrev = data;
	struct stSyncTmr *synctmr = synctmrev->synctmr;
	struct stSyncTmrFunc *synctmrfunc;
	GList *firstelem;

	firstelem = g_list_first(synctmr->events);
	synctmrfunc = (struct stSyncTmrFunc *) firstelem->data;
	synctmr->events = g_list_remove_link(synctmr->events, firstelem);
	g_list_free_1(firstelem);
	
	g_hash_table_remove(synctmr->timeouts,
			    GUINT_TO_POINTER(synctmrev->
					     timeout_handler_id));
	g_free(synctmrev);
	
	/* run timeout, ignore return value (we don't repeat timeouts) */
	(*synctmrfunc->function) (synctmrfunc->data);
	g_free(synctmrfunc);
	return FALSE;
}

void synctmr_timeout_add(struct stSyncTmr *synctmr, guint32 interval,
			 GtkFunction function, gpointer data)
{
	struct stSyncTmrEvent *synctmrev;
	struct stSyncTmrFunc *synctmrfunc;

	synctmrfunc = g_new0(struct stSyncTmrFunc, 1);
	synctmrfunc->function = function;
	synctmrfunc->data = data;
	synctmr->events = g_list_append(synctmr->events, synctmrfunc);

	synctmrev = g_new0(struct stSyncTmrEvent, 1);
	synctmrev->synctmr = synctmr;
	synctmrev->timeout_handler_id =
	    gtk_timeout_add(interval, timeout_happened, synctmrev);
	g_hash_table_insert(synctmrev->synctmr->timeouts,
			    GUINT_TO_POINTER(synctmrev->
					     timeout_handler_id),
			    synctmrev);
}
